% Benchmark case for gravity driven poiseuille flow with second order 
% boundary conditions by calculating the stress tensor S (BC4 in Latt 2008). 

% ######################################################################
% ##################### INITIALIZATION #################################
% ######################################################################
clear
NY=input('Number of cells in crossflow direction (NY):');
NX=NY;
h=NY-1;
delta_y=1/h;
u_lattice=1/h; %Or: u_L~delta y. delta y=1/h where h=(NY-1). This gives omega=const.
umax=1; %Macroscopic velocity
%uavg=(2/3)*umax;
Re=10;
nu=u_lattice*h/Re; 
tau=3*nu+0.5;
omega=1/tau;
% omega=1.0; %Relaxation frequency
% tau=1/omega; %Relaxation time
% nu=(1/3)*(tau-0.5); %Kinematic viscosity
% Re=4.4; %Reynolds number
% umax=0.1;
% uavg=2/3*umax;
% NY=round(Re*nu/uavg); %Lattice units in y-direction. Re is based on the average velocity. Min number of points
% NX=NY;
i=1:NX; %Computational domain:x-count
j=1:NY; %Computational domain:y-count
J=2:NY-1;%Wet nodes i y-direction
jx=zeros(NY,NX); %x-momentum
jy=zeros(NY,NX); %y-momentum
rho=ones(NY,NX); %Density
f=zeros(NY,NX); %Particle distribution function. f=f_{in}
feq=zeros(NY,NX,9);
fneq=zeros(NY,NX,9);
fout=zeros(NY,NX);
g=8*u_lattice*nu/h^2; %Gravity. 
G=g*rho; %Forcing term
Gy=zeros(NY,NX);
Z=zeros(NY,NX,9); %Force term in collision. Eqn.(2.74) in Latt
Q=zeros(2,2,9);
S=zeros(2,2); %Rate of stress tensor
%Analytical solution
C1=(4*umax/h^3)*(NY^2-1);
C2=4*umax/h^2 - C1;
uana=-(4*umax/h^2)*j.^2 + C1*j + C2;
%Other input variables
aa=1-0.5*omega;

%D2Q9 lattice constants
t0=4/9; t1=1/9; t2=1/36;
t=[t1,t1,t1,t1,t2,t2,t2,t2,t0]; %Lattice weights
f0=1; f1=3; f2=4.5; f3=1.5;
cx=[1,0,-1,0,1,-1,-1,1,0];
cy=[0,1,0,-1,1,1,-1,-1,0];
q=9; %number of velocities on D2Q9 lattice


%        C6  C2  C5       ^ y
%          \ | /          |
%         C3-C0-C1        | 
%          / | \          | 
%        C7  C4  C8         -----> x


ux=jx./rho; %Compute macroscopic x-velocity
uy=jy./rho; %Compute macroscopic y-velocity

uxsq=ux.*ux;
uysq=uy.*uy;

usq=uxsq + uysq;

uxuy5=ux + uy;
uxuy6=-ux + uy;
uxuy7=-ux - uy;
uxuy8=ux - uy;

% Initialize the particle distribution function to equilibrium: f(t=0)=f^{eq} 
f(j,i,1)=t(1)*rho(j,i).*(f0 + f1.*ux(j,i)  + f2.*uxsq(j,i) -          f3*usq(j,i));
f(j,i,2)=t(2)*rho(j,i).*(f0 + f1.*uy(j,i)  + f2.*uysq(j,i) -          f3*usq(j,i));
f(j,i,3)=t(3)*rho(j,i).*(f0 - f1.*ux(j,i)  + f2.*uxsq(j,i) -          f3*usq(j,i));
f(j,i,4)=t(4)*rho(j,i).*(f0 - f1.*uy(j,i)  + f2.*uysq(j,i) -          f3*usq(j,i));
f(j,i,5)=t(5)*rho(j,i).*(f0 + f1.*uxuy5(j,i) + f2.*uxuy5(j,i).*uxuy5(j,i)-    f3*usq(j,i));
f(j,i,6)=t(6)*rho(j,i).*(f0 + f1.*uxuy6(j,i) + f2.*uxuy6(j,i).*uxuy6(j,i)-    f3*usq(j,i));
f(j,i,7)=t(7)*rho(j,i).*(f0 + f1.*uxuy7(j,i) + f2.*uxuy7(j,i).*uxuy7(j,i)-    f3*usq(j,i));
f(j,i,8)=t(8)*rho(j,i).*(f0 + f1.*uxuy8(j,i) + f2.*uxuy8(j,i).*uxuy8(j,i)-    f3*usq(j,i));
f(j,i,9)=t(9)*rho(j,i).*(f0 -                                       f3*usq(j,i));

% Defining Q=c_i*c_i-c_s^2*eye
for ii=1:q;
    for a=1:2;
        for b=1:2;
            if a==1
                if b==1
                    Q(b,a,ii)=cx(ii)*cx(ii)-1/3;
                else
                    Q(b,a,ii)=cx(ii)*cy(ii);
                end
            else
                if b==1
                    Q(b,a,ii)=cy(ii)*cx(ii);
                else
                    Q(b,a,ii)=cy(ii)*cy(ii)-1/3;
                end
            end
        end
    end
end

Qxx=Q(1,1,:);Qxy=Q(1,2,:);
Qyx=Q(2,1,:);Qyy=Q(2,2,:);
Sxx_l=zeros(1,NX);Syy_l=zeros(1,NX);Sxx_u=zeros(1,NX);Syy_u=zeros(1,NX); 
Sxy_l=zeros(1,NX); Sxy_u=zeros(1,NX); %If S=0 tensor. This is the equilibrium BC

% ##################### INITIALIZATION DONE ############################
% ######################################################################

fout=f;

difference=1;
n=1;
convergence_criterion=1e-12;
%COLLISION PHASE
err=ones(n,1);
maxitr=40000; %Maximum number of iterations
% for n=1:maxitr

% ######################################################################
% ##################### LOOP STARTS HERE!! ############################
% ######################################################################

tic;
while difference(n)*NY>convergence_criterion && n<maxitr    
    
%Update the values of rho, jx and jy    
% rho(1,:)=6*(f(1,:,2)+f(1,:,5)+f(1,:,6));
% rho(NY,:)=6*(f(NY,:,4)+f(NY,:,7)+f(NY,:,8));

rho(1,:)=6*(f(1,:,4)+f(1,:,7)+f(1,:,8));
rho(NY,:)=6*(f(NY,:,2)+f(NY,:,5)+f(NY,:,6));

rho(J,:)=f(J,:,1)+f(J,:,2)+f(J,:,3)+f(J,:,4)+f(J,:,5)+f(J,:,6)+f(J,:,7)+f(J,:,8)+f(J,:,9);

jx(J,i)=cx(1)*f(J,i,1)+cx(2)*f(J,i,2)+cx(3)*f(J,i,3)+cx(4)*f(J,i,4)+cx(5)*f(J,i,5)+cx(6)*f(J,i,6)+cx(7)*f(J,i,7)+cx(8)*f(J,i,8);
jy(J,i)=cy(1)*f(J,i,1)+cy(2)*f(J,i,2)+cy(3)*f(J,i,3)+cy(4)*f(J,i,4)+cy(5)*f(J,i,5)+cy(6)*f(J,i,6)+cy(7)*f(J,i,7)+cy(8)*f(J,i,8);

G=rho.*g; %Forcing term

ux=jx./rho; 
uy=jy./rho;

ux=ux+0.5*G; %Eqn. (2.30) in Latt

ux(1,:)=0; %No-slip at boundaries
ux(NY,:)=0; %No-slip at boundaries
uy(1,:)=0; %No-slip at boundaries
uy(NY,:)=0; %No-slip at boundaries

ueqx=ux; %The velocities used in f^{eq} depend on the equilibrium velocities.
ueqy=uy;

uxsq=ueqx.*ueqx;
uysq=ueqy.*ueqy;

usq=uxsq + uysq;

uxuy5=ux + uy;
uxuy6=-ux + uy;
uxuy7=-ux - uy;
uxuy8=ux - uy;

% Compute the equilibrium function f^{eq}
% All cells (different from bounceback boundary condition)
feq(j,i,1)=t(1)*rho(j,i).*(f0 + f1.*ueqx(j,i)  + f2.*uxsq(j,i) -                 f3.*usq(j,i));
feq(j,i,2)=t(2)*rho(j,i).*(f0 + f1.*ueqy(j,i)  + f2.*uysq(j,i) -                 f3.*usq(j,i));
feq(j,i,3)=t(3)*rho(j,i).*(f0 - f1.*ueqx(j,i)  + f2.*uxsq(j,i) -                 f3.*usq(j,i));
feq(j,i,4)=t(4)*rho(j,i).*(f0 - f1.*ueqy(j,i)  + f2.*uysq(j,i) -                 f3.*usq(j,i));
feq(j,i,5)=t(5)*rho(j,i).*(f0 + f1.*uxuy5(j,i) + f2.*uxuy5(j,i).*uxuy5(j,i)-     f3.*usq(j,i));
feq(j,i,6)=t(6)*rho(j,i).*(f0 + f1.*uxuy6(j,i) + f2.*uxuy6(j,i).*uxuy6(j,i)-     f3.*usq(j,i));
feq(j,i,7)=t(7)*rho(j,i).*(f0 + f1.*uxuy7(j,i) + f2.*uxuy7(j,i).*uxuy7(j,i)-     f3.*usq(j,i));
feq(j,i,8)=t(8)*rho(j,i).*(f0 + f1.*uxuy8(j,i) + f2.*uxuy8(j,i).*uxuy8(j,i)-     f3.*usq(j,i));
feq(j,i,9)=t(9)*rho(j,i).*(f0 -                                                  f3.*usq(j,i));

% G=rho.*g;
%Z(j,i,a) is eqn.(2.74) in Latt
for k=1:9
    p(:,:,k)=(cx(k)*ones(NY,NX)-ux).*G + (cy(k)*ones(NY,NX)-uy).*Gy; %G=Gx
    qq(:,:,k)=(cx(k)*ux*cx(k)).*G + (cy(k)*uy*cx(k)).*G + (cx(k)*ux*cy(k)).*Gy + (cy(k)*uy*cy(k)).*Gy;
    Z(:,:,k)=aa*t(k)*3*p(:,:,k) + aa*t(k)*9*qq(:,:,k);
end


% #############################################################
% ################### COLLISION PHASE #########################
% #############################################################

%Computing the rate of stress tensor S
dudy_l=1.5*ux(NY,:)-2*ux(NY-1,:)+0.5*ux(NY-2,:);
Sxy_l=0.5*dudy_l;Syx_l=Sxy_l; %Activate for S \neq 0

dudy_u=-1.5*ux(1,:)+2*ux(2,:)-0.5*ux(3,:);
Sxy_u=0.5*dudy_u;Syx_u=Sxy_u; %Activate for S \neq 0

% Q_inner_S_l(a)=trace(Qxy(a)'*Sxy_l);
for c=1:q;
Q_inner_S_l(c,:)=Qxx(c)*Sxx_l + 2*Qxy(c)*Sxy_l + Qyy(c)*Syy_l;
Q_inner_S_u(c,:)=Qxx(c)*Sxx_u + 2*Qxy(c)*Sxy_u + Qyy(c)*Syy_u;
fneq(NY,:,c)=-(3*t(c)/omega)*rho(NY,:).*Q_inner_S_l(c,:);
fneq(1,:,c)=-(3*t(c)/omega)*rho(1,:).*Q_inner_S_u(c,:);
end

f(1,:,:)=feq(1,:,:)+fneq(1,:,:); %Reconstructing the populations (16)
f(NY,:,:)=feq(NY,:,:)+fneq(NY,:,:); %Reconstructing the populations (16)
Z(1,:,:)=0;Z(NY,:,:)=0; %No forcing on boundaries
OMEGA=-omega*(f-feq) + Z;

% ################ Collision step #############################
  fout=f + OMEGA;   %Relaxation towards local equilibrium
% #############################################################



% #############################################################
% ####################### STREAMING PHASE #####################
% #############################################################
 
% %################ "Old" streaming #############################
% 
% f(j,2:NX,1)            = fout(j,1:NX-1,1); %Propagation of c1
% f(2:NY,i,2)            = fout(1:NY-1,i,2); %Propagation of c2
% f(j,1:NX-1,3)          = fout(j,2:NX,3);   %Propagation of c3
% f(1:NY-1,i,4)          = fout(2:NY,i,4);   %Propagation of c4
% f(2:NY,2:NX,5)         = fout(1:NY-1,1:NX-1,5);%Propagation of c5
% f(2:NY,1:NX-1,6)       = fout(1:NY-1,2:NX,6);%Propagation of c6
% f(1:NY-1,1:NX-1,7)     = fout(2:NY,2:NX,7);%Propagation of c7
% f(1:NY-1,2:NX,8)       = fout(2:NY,1:NX-1,8);%Propagation of c8
% f(j,i,9)               = fout(j,i,9);%Propagation of c0. Necessary?
% 
% % ################################################################
% % ############ PERIODIC BOUNDARY CONDITIONS ######################
% % ################################################################
% f(j,1,1)        =   fout(j,NX,1);
% f(j,NX,3)       =   fout(j,1,3);
% f(2:NY,1,5)     =   fout(1:NY-1,NX,5);
% f(2:NY,NX,6)    =   fout(1:NY-1,1,6);
% f(1:NY-1,NX,7)  =   fout(2:NY,1,7);
% f(1:NY-1,1,8)   =   fout(2:NY,NX,8);

% 4 fun!!
for lll=1:q
    f(:,:,lll)=circshift(fout(:,:,lll), [cy(lll),cx(lll),0]);
end

n=n+1;
uprofile=jx(:,ceil(NY/2))./rho(:,ceil(NY/2))/u_lattice;
vprofile=jy(:,ceil(NY/2))./rho(:,ceil(NY/2))/u_lattice;
% figure(1)
% plot(uprofile,'k+')
time(n)=delta_y*n;
xmomentum(n)=sum(jx(:,ceil(NY/2)));
errtemp=sqrt((uprofile(j)-uana(j)').^2+(vprofile(j)).^2);
err(n)=(1/NY)*sum(errtemp);
difference(n)=abs(err(n-1)-err(n)); %NB! abs()
difference(n-1)=difference(n);
% pause(.1)
end
simtime=toc;
N=1:n;
disp('#### RESULTS FROM POISEUILLE_JLFD.M ####')
% ##################################################################
% ######################## LOOP ENDS ###############################
% ########################  PLOTTING ###############################
% ##################################################################
figure(1)
plot(uprofile,'k+')
hold on;
plot(uana,'k-')
xlabel('Grid node NY')
title(['Poiseuille flow with NY=',num2str(NY),' nodes'])
ylabel('u(NY)/ u_{max}')
legend('u(y)','u_{ana}')

figure(2)
plot(time,xmomentum,'k.')
xlabel('Time t')
ylabel('x-momentum j_x(t)')

% hold on
% figure(3)
% loglog(NY,err(n),'k*')
% grid minor
% title('2-norm of error for gravity driven Poiseuille flow')
% xlabel('NY')
% ylabel('Error \epsilon')
% axis([10^1 10^2 10^(-3) 10^0])

disp(['Number of iterations required was ',num2str(n-1)])
disp(['Maximum velocity= ',num2str(max(uprofile))])
disp(['Error for NY= ',num2str(NY), 'nodes is ',num2str(err(n))])
disp(['iterationtime w/',num2str(n-1),' iterations was ',num2str(simtime)])

% hold off
% figure(4)
% semilogy(2:n,difference(2:n),'k.')
% title('Difference in error between iteration n and (n-1)')
% legend(['NY=',num2str(NY),' nodes'])
% xlabel('Iteration n')
% ylabel('Difference')
% % =\epsilon_n-\epsilon_{n-1}')







